#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Tokens */
#define TK_IDENTIFIER           1
#define TK_INTEGER              2
#define TK_FLOAT                3
#define TK_STRING               4
#define TK_CHAR                 5
#define TK_SEMICOLON            6
#define TK_CURLY_OPEN           7
#define TK_CURLY_CLOSE          8
#define TK_PARENTHESIS_OPEN     9
#define TK_PARENTHESIS_CLOSE    10
#define TK_SQUARE_OPEN          11
#define TK_SQUARE_CLOSE         12
#define TK_COMMA                13
#define TK_DOT                  14
#define TK_EQUAL_ASSIGN         15
#define TK_EQUALS               16
#define TK_LT                   17
#define TK_LTE                  18
#define TK_GT                   19
#define TK_GTE                  20
#define TK_PLUS                 21
#define TK_PLUS_PLUS            22
#define TK_MINUS                23
#define TK_MINUS_MINUS          24
#define TK_MUL                  25
#define TK_DIV                  26
#define TK_MODULO               27
#define TK_BANG                 28
#define TK_QMARK                29
#define TK_COLON                30
#define TK_AND                  31
#define TK_AND_AND              32
#define TK_OR                   33
#define TK_OR_OR                34
#define TK_NOT_EQUAL            35
#define TK_KEYWORD_INT          36
#define TK_KEYWORD_FLOAT        37
#define TK_KEYWORD_DOUBLE       38
#define TK_KEYWORD_BOOL         39
#define TK_KEYWORD_STRING       40
#define TK_KEYWORD_WHILE        41
#define TK_KEYWORD_UNTIL        42
#define TK_KEYWORD_IF           43
#define TK_KEYWORD_ELSE         44
#define TK_KEYWORD_TRUE         45
#define TK_KEYWORD_FALSE        46
#define TK_KEYWORD_NAMESPACE    47
#define TK_KEYWORD_CONSOLE      48
#define TK_KEYWORD_WRITELINE    49 
#define TK_KEYWORD_STATIC       50
#define TK_KEYWORD_FUNC_NAME    51
#define TK_KEYWORD_USING        52
#define TK_KEYWORD_SYSTEM       53
#define TK_KEYWORD_CLASS        54
#define TK_KEYWORD_CHAR         55
#define TK_KEYWORD_VOID         56
#define TK_MAIN                 57

/* Aliased data-types */
#define TOKEN                   int
#define STATE                   int

/* Lengths */
#define N_LEN                   256  // Buffer is 2N size
#define LEXEME_LEN              512
#define TKNLXM_LEN              3    // Minimum 3, since we need 3 in the buffer for negative number detection
                                     // by using a lookahead

/* Reused conditional fragments */
#define RE_NUM                  ch >= '0' && ch <= '9'
#define RE_ID                   (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || ch == '_' || ch == '@'

/* Reserved Keywords */
#define RES_KEY                 "int", "float", "double", "bool", "string", "while", "until", "if", "else", "true", "false", "namespace", "Console", "WriteLine", "static", "func_name", "using", "System", "class", "char","void","Main"

typedef struct TokenLexeme {
    TOKEN token;
    char *lexeme;
    int line;
} TokenLexeme;

TokenLexeme *tl_delim = NULL;   // Used to return TokenLexeme for delimiters (eg. identifiers are delimited by operators)
                                // on next call of get_next_token(), even though
                                // the delimiter char is already read

FILE *src_file = NULL;          // fopen'd r-mode pointer to source-code file
char buf[2*N_LEN];              // We buffer the file-stream into memory using a 2-buffer scheme
char *cptr = NULL;              // Pointer into buf
STATE state;                    // Keeps track of state in the current FSM being executed. Made global
                                // so that we can link machines to each other at arbitrary states.
char *lexeme;                   // Current lexeme being read
int lexeme_i;                   // Index of the last character of the current lexeme being read

short read_to_buf = 1;          // Used as a flag to not re-read from file if a _lexer_rewind() was called
TokenLexeme *tlbuf[TKNLXM_LEN]; // Buffer for TokenLexemes to be returned (used for token-lookaheads)

int lexer_error_count = 0;      // Used to keep count of lexing errors so far
int lexer_line_count = 0;       // Used to keep count of the line while reporting tokens and errors


void lexer(FILE *src) {
    src_file = src;
    cptr = buf;
}

// Implements 2-buffer lexing scheme
char _getbufc() {
    int c = N_LEN;
    if (cptr == buf + (N_LEN)*sizeof(char)) {
        if (read_to_buf) {
            c = fread(cptr, sizeof(char), N_LEN, src_file);
        } else {
            read_to_buf = 1;
        }
    } else if (cptr == buf + (2*N_LEN)*sizeof(char)) {
        cptr = buf;
        if (read_to_buf) {
            c = fread(cptr, sizeof(char), N_LEN, src_file);
        } else {
            read_to_buf = 1;
        }
    } else if (cptr == buf) {
        if (read_to_buf) {
            c = fread(cptr, sizeof(char), N_LEN, src_file);
        } else {
            read_to_buf = 1;
        }
    }

    if (c < N_LEN) {
        cptr[c] = '\0';
    }

    char ret = *cptr;
    if (ret == '\0') {
        // File is over!
        cptr = NULL;
    } else {
        cptr++;
    }

    return ret;
}

// Rewinds one character that was last read
void _lexer_rewind() {
    if (cptr == buf + N_LEN + 1) {
        // Buffer 2 was just loaded and first char read
        read_to_buf = 0;
        cptr--;
    } else if (cptr == buf + 1) {
        // Buffer 1 was just loaded and first char read
        read_to_buf = 0;
        cptr--;
    } else {
        cptr--;
    }
}

// Returns char returned by _getbufc()
// but also writes it into str[pos]
char _getbufc_to_str(char *str, int pos) {
    return (str[pos] = _getbufc());
}

// Checks whether current token is being delimited by
// new token
int _is_delim(char ch) {
    if (ch == '\n' || ch == '\t' || ch == ' ' || ch == '\0') {
        if (ch == '\n') {
            lexer_line_count++;
        }

        return 1;
    } else if (ch == ';' || ch == '{' || ch == '}' || ch == '(' ||
               ch == ')' || ch == '[' || ch == ']' || ch == ',' ||
               ch == '+' || ch == '-' || ch == '/' || ch == '*' ||
               ch == '%' || ch == '&' || ch == '|' || ch == '!' ||
               ch == '?' || ch == ':' || ch == '=' || ch == '.' ||
               ch == '<' || ch == '>') {
        tl_delim = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
        tl_delim->lexeme = calloc(2, sizeof(char));
        tl_delim->lexeme[0] = ch;
        tl_delim->line = lexer_line_count;

        char ch2; // In case we need to peek onto the next char!
        
        switch (ch) {
            case ';':
                tl_delim->token = TK_SEMICOLON;
                break;
            case '{':
                tl_delim->token = TK_CURLY_OPEN;
                break;
            case '}':
                tl_delim->token = TK_CURLY_CLOSE;
                break;
            case '(':
                tl_delim->token = TK_PARENTHESIS_OPEN;
                break;
            case ')':
                tl_delim->token = TK_PARENTHESIS_CLOSE;
                break;
            case '[':
                tl_delim->token = TK_SQUARE_OPEN;
                break;
            case ']':
                tl_delim->token = TK_SQUARE_CLOSE;
                break;
            case ',':
                tl_delim->token = TK_COMMA;
                break;
            case '.':
                tl_delim->token = TK_DOT;
                break;
            case '*':
                tl_delim->token = TK_MUL;
                break;
            case '%':
                tl_delim->token = TK_MODULO;
                break;
            case '?':
                tl_delim->token = TK_QMARK;
                break;
            case ':':
                tl_delim->token = TK_COLON;
                break;
            case '+':
                tl_delim->token = TK_PLUS;
                ch2 = _getbufc();

                switch (ch2) {
                    case '+':
                        tl_delim->lexeme[1] = '+';
                        tl_delim->token = TK_PLUS_PLUS;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '-':
                tl_delim->token = TK_MINUS;
                ch2 = _getbufc();

                switch (ch2) {
                    case '-':
                        tl_delim->lexeme[1] = '-';
                        tl_delim->token = TK_MINUS_MINUS;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '<':
                tl_delim->token = TK_LT;
                ch2 = _getbufc();

                switch (ch2) {
                    case '=':
                        tl_delim->lexeme[1] = '=';
                        tl_delim->token = TK_LTE;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '>':
                tl_delim->token = TK_GT;
                ch2 = _getbufc();

                switch (ch2) {
                    case '=':
                        tl_delim->lexeme[1] = '=';
                        tl_delim->token = TK_GTE;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '!':
                tl_delim->token = TK_BANG;
                ch2 = _getbufc();

                switch (ch2) {
                    case '=':
                        tl_delim->lexeme[1] = '=';
                        tl_delim->token = TK_NOT_EQUAL;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '=':
                tl_delim->token = TK_EQUAL_ASSIGN;
                ch2 = _getbufc();

                switch (ch2) {
                    case '=':
                        tl_delim->lexeme[1] = '=';
                        tl_delim->token = TK_EQUALS;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '&':
                tl_delim->token = TK_AND;
                ch2 = _getbufc();

                switch (ch2) {
                    case '&':
                        tl_delim->lexeme[1] = '&';
                        tl_delim->token = TK_AND_AND;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '|':
                tl_delim->token = TK_OR;
                ch2 = _getbufc();

                switch (ch2) {
                    case '|':
                        tl_delim->lexeme[1] = '|';
                        tl_delim->token = TK_OR_OR;
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
            case '/':
                tl_delim->token = TK_DIV;
                ch2 = _getbufc();

                switch (ch2) {
                    case '/':                   // Comment-removal!
                        while (_getbufc() != '\n');
                        tl_delim = NULL;
                        break;
                    case '*':
                        while (1) {
                            char ch3 = _getbufc();
                            if (ch3 == '*') {
                                char ch4;
                                if ((ch4 = _getbufc()) == '/') {
                                    tl_delim = NULL;
                                    break;
                                } else if (ch4 == '*') {
                                    _lexer_rewind();
                                }
                            }
                        }
                        break;
                    default:
                        _lexer_rewind();
                        break;
                }

                break;
        }
        return 1;
    }

    return 0;
}

int _keyword_tok_id(char *str) {
    char *res_key_arr[] = {RES_KEY};

    for (int i = 0, len = sizeof(res_key_arr)/sizeof(char **); 
         i < len;
         i++) {

        if (strcmp(lexeme, res_key_arr[i]) == 0) {
            return i;
        }
    }

    return -1;
}

/* Our automata magic in da' house! */

TokenLexeme* _fsm_numeric() {
    state = 0;

    while (1) {
        char ch = _getbufc_to_str(lexeme, lexeme_i++);
        switch (state) {
            case 0:
                if (ch == '.') {
                    // We have a floating point number on our hands!
                    state = 2;
                } else if (RE_NUM) {
                    state = 0;
                } else if (_is_delim(ch)) {
                    // Accepted as integer literal!
                    lexeme[lexeme_i - 1] = '\0';
                    TokenLexeme *ret = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
                    ret->token = TK_INTEGER;
                    ret->lexeme = lexeme;
                    ret->line = lexer_line_count;
                    return ret;
                } else {
                    state = 1;
                }
                break;
            case 1:
                if (_is_delim(ch)) {
                    lexeme[lexeme_i - 1] = '\0';
                    printf("Line: %d | Invalid numeric token: %s\n", lexer_line_count, lexeme);
                    lexer_error_count++;
                    return NULL;
                    break;
                }
            case 2:
                if (RE_NUM) {
                    state = 2;
                } else if (_is_delim(ch)) {
                    // Accepted as floating literal!
                    lexeme[lexeme_i - 1] = '\0';
                    TokenLexeme *ret = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
                    ret->token = TK_FLOAT;
                    ret->lexeme = lexeme;
                    ret->line = lexer_line_count;
                    return ret;
                } else {
                    state = 1;
                }
                break;
        }
    }

    return NULL;
}

TokenLexeme* _fsm_id() {
    state = 0;

    while (1) {
        char ch = _getbufc_to_str(lexeme, lexeme_i++);
        switch (state) {
            case 0:
                if (RE_ID || (RE_NUM)) {
                    state = 0;
                } else if (_is_delim(ch)) {
                    lexeme[lexeme_i - 1] = '\0';
                    TokenLexeme *ret = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
                    ret->lexeme = lexeme;
                    ret->line = lexer_line_count;

                    // Check if our identifier is not actually a reserved keyword
                    int kw_token;
                    ret->token = (kw_token = _keyword_tok_id(lexeme)) >= 0 ? TK_KEYWORD_INT + kw_token : TK_IDENTIFIER;
                    return ret;
                } else {
                    state = 1;
                }
                break;
            case 1:
                lexeme[lexeme_i - 1] = '\0';
                printf("Line: %d | Invalid identifier token: %s\n", lexer_line_count, lexeme);
                lexer_error_count++;
                return NULL;
                break;
        }
    }
}

TokenLexeme* _fsm_string() {
    while (1) {
        char ch = _getbufc_to_str(lexeme, lexeme_i++);
        switch (state) {
            case 0:
                switch (ch) {
                    case '\\': // That's actually just one \, please note!
                        state = 1;
                        break;
                    case '"':
                        // Accepted string literal!
                        lexeme[lexeme_i] = '\0';
                        TokenLexeme *ret = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
                        ret->lexeme = lexeme;
                        ret->token = TK_STRING;
                        ret->line = lexer_line_count;
                        return ret;
                        break;
                    default:
                        break;
                }
                break;
            case 1:
                switch (ch) {
                    case 't': case 'n': case 'r': case '"': case '\\':
                        // Alrighty then!
                        state = 0;
                        break;
                    default:
                        printf("Line: %d | Illegal escape sequence in string literal: \\%c\n", lexer_line_count, ch);
                        lexer_error_count++;
                        return NULL;
                        break;
                }
                break;
        }
    }

    return NULL;
}

TokenLexeme* _fsm_char() {
    state = 0;

    while (1) {
        char ch = _getbufc_to_str(lexeme, lexeme_i++);
        switch (state) {
            case 0:
                switch (ch) {
                    case '\\':
                        state = 1;
                        break;
                    case '\'':
                        printf("Line: %d | Empty character literal\n", lexer_line_count);
                        return NULL;
                        break;
                    default:
                        state = 2;
                        break;
                }
                break;
            case 1:
                switch (ch) {
                    case 't': case 'n': case 'r': case '"': case '\\': case '\'':
                        state = 2;
                        break;
                    default:
                        printf("Line: %d | Invalid escape sequence: %s\n", lexer_line_count, lexeme);
                        return NULL;
                        break;
                }
                break;
            case 2:
                switch (ch) {
                    case '\'':
                        // Accepted!
                        lexeme[lexeme_i] = '\0';
                        TokenLexeme *ret = (TokenLexeme *) calloc(1, sizeof(TokenLexeme));
                        ret->token = TK_CHAR;
                        ret->lexeme = lexeme;
                        ret->line = lexer_line_count;
                        return ret;
                        break;
                    default:
                        printf("Line: %d | Illegal character literal: %s\n", lexer_line_count, lexeme);
                        return NULL;
                        break;
                }
                break;
        }
    }

    return NULL;
}

// Checks if any more raw tokens are to
// be read
int _has_more_tokens() {
    return cptr != NULL;
}

// Nice version of _has_more_tokens()
// Checks if no more tokens are in tlbuf[]
// to return
int has_more_tokens() {
    return _has_more_tokens() || tlbuf[0] != NULL;
}

// Raw token return, returns NULL when token
// is whitespace as well
TokenLexeme* _get_next_token() {
    if (tl_delim == NULL) {
        lexeme = calloc(LEXEME_LEN, sizeof(char));
        lexeme_i = 0;

        state = 0;

        char ch = _getbufc_to_str(lexeme, lexeme_i++);

        switch (state) {
            case 0:
                if (RE_NUM) {
                    return _fsm_numeric();
                } else if (RE_ID) {
                    // We'll check if it's a keyword inside _fsm_id()
                    return _fsm_id();
                } else if (ch == '"') {
                    return _fsm_string();
                } else if (ch == '+' || ch == '-' || ch == '*' || ch == '/' || ch == '.' ||
                           ch == '&' || ch == '|' || ch == '=' || ch == '%' || ch == '<' ||
                           ch == '>') {
                    if (_is_delim(ch)) {
                        TokenLexeme *ret = tl_delim;
                        tl_delim = NULL;
                        return ret;
                    }
                } else if (ch == '\'') {
                    return _fsm_char();
                } else if (_is_delim(ch)) {
                    TokenLexeme *ret = tl_delim;
                    tl_delim = NULL;
                    return ret;
                }

                break;
        }

        return NULL;
    } else {
        TokenLexeme *ret = tl_delim;
        tl_delim = NULL;
        return ret;
    }
}

// Shift the entire buffer by one to the left
void _shift_tlbuf_to_left() {
    for (int i = 0; i < TKNLXM_LEN - 1; i++) {
        tlbuf[i] = tlbuf[i + 1];
    }

    // Fill up the last NULL created if possible
    if (_has_more_tokens()) {
        short tok_found = 0;
        while (_has_more_tokens()) {
            TokenLexeme *tl;
            if ((tl = _get_next_token()) != NULL) {
                tlbuf[TKNLXM_LEN - 1] = tl;
                tok_found = 1;
                break;
            }
        }

        if (!tok_found) {
            tlbuf[TKNLXM_LEN - 1] = NULL;
        }
    } else {
        tlbuf[TKNLXM_LEN - 1] = NULL;
    }
}

// Nice, cleaned up version of _get_next_token()
// Skips whitespace and stuff
TokenLexeme* get_next_token() {
    // Runs the first time to populate buffer at the start
    int tlbuf_i = 0;
    while (tlbuf_i < TKNLXM_LEN && tlbuf[tlbuf_i] == NULL && _has_more_tokens()) {
        TokenLexeme *tl;
        if ((tl = _get_next_token()) != NULL) {
            tlbuf[tlbuf_i++] = tl;
        }
    }

    // To return the top of the buffer
    TokenLexeme *ret = tlbuf[0];

    _shift_tlbuf_to_left();

    // Do we have a sign-explicit number in store?
    if (ret->token >= TK_SEMICOLON && ret->token <= TK_NOT_EQUAL) {
        // Delimiter spotted
        TokenLexeme *t1 = tlbuf[0];
        if (t1 != NULL && (t1->token == TK_PLUS || t1->token == TK_MINUS)) {
            TokenLexeme *t2 = tlbuf[1];
            if (t2 != NULL && (t2->token == TK_INTEGER || t2->token == TK_FLOAT)) {
                // Append the + or - to the INT or FLOAT literal
                int len = strlen(t2->lexeme);
                t2->lexeme[len + 1] = '\0';
                for (int i = len; i > 0; i--) {
                    t2->lexeme[i] = t2->lexeme[i - 1];
                }

                switch (t1->token) {
                    case TK_PLUS:
                        t2->lexeme[0] = '+';
                        break;
                    case TK_MINUS:
                        t2->lexeme[0] = '-';
                        break;
                }

                // Delete the redundant + or - token and shift buffer hence
                // to left
                _shift_tlbuf_to_left();
            }
        }
    }

    return ret;
}
